"""
Factory to wrap Pimoroni sensor objects, many of which have different ways to read data and structure the results. Because this is just a wrapper class, you will need to separately install the Pimoroni library for each sensor you want to poll.

Example for `the MICS6814 3-in-1 Gas Sensor <https://thepihut.com/products/mics6814-3-in-1-gas-sensor-breakout-co-no2-nh3>`_: ::

   # install the sensor library
   pip3 install pimoroni-mics6814

   # create the sensor to poll
   GenericSensor.mics6814_i2c()

Warning: sensors like ``mics6814_i2c`` rely on `the Pimoroni IO Expander library <https://github.com/pimoroni/ioe-python>`_, which uses `smbus2 <https://github.com/kplindegaard/smbus2>`_ internally and doesn't implement any kind of locking. This means it's possible for multiple sensors to conflict if they poll for data around the exact same time.
"""

from snsary.models import Reading
from snsary.sources import PollingSensor
from snsary.utils import scraper


class GenericSensor(PollingSensor):
    @classmethod
    def mics6814_i2c(cls):
        """
        Returns a sensor configured to scrape ``oxidising``, ``reducing`` and ``nh3`` values in Ohms. Calling this function also switches off the LED of the MICS 6814, which is on by default.

        Requires `the mics6814-python library <https://github.com/pimoroni/mics6814-python>`_. Note that, while Pimoroni do provide `a library for CircuitPython <https://github.com/pimoroni/Pimoroni_CircuitPython_MICS6814>`_ (similarly to many Adafruit sensors), this only supports pin-based wiring.
        """

        # data is stored as slots in reading
        from mics6814 import Mics6814Reading

        class_scraper = scraper.for_class(Mics6814Reading)

        # create once to avoid descriptor leak
        from mics6814 import MICS6814

        instance = MICS6814()

        # turn off bright LED (on by default)
        instance.set_led(0, 0, 0)

        def read_fn():
            scraps = class_scraper(instance.read_all())
            # filter out internal "adc" resistance values
            return filter(lambda scrap: scrap[0] != "adc", scraps)

        return cls(name="MICS6814", read_fn=read_fn)

    def __init__(self, *, name, read_fn):
        PollingSensor.__init__(self, period_seconds=10)
        self.__name = name
        self.__read_fn = read_fn

    @property
    def name(self):
        return self.__name

    def sample(self, timestamp, **kwargs):
        return [
            Reading(
                sensor_name=self.name,
                name=name,
                value=value,
                timestamp=timestamp,
            )
            for name, value in self.__read_fn()
        ]



"""
Maps each :mod:`Reading <snsary.models.reading>` to a new one with the name altered as specified. This can be useful to distinguish the same :mod:`Readings <snsary.models.reading>` being aggregated in different ways.

Supported alterations include:

    - ``to`` - replaces the name of the reading with the one specified.
    - ``append`` - adds to the new or existing name of the reading.
"""

from .function import Function


class Rename(Function):
    def __init__(self, append="", to=None):
        self.__append = append
        self.__to = to

    def __call__(self, reading):
        new_name = self.__to if self.__to else reading.name
        new_name += self.__append
        return [reading.dup(name=new_name)]



"""
Wrapper for most Adafruit sensor objects, which usually have one or more properties to retrieve the current values of the sensor. GenericSensor polls a given sensor object and publishes the latest value of each property as a :mod:`Reading <snsary.models.reading>`.

Most Adafruit sensors come with a driver written for CircuitPython, which can be used in pure Python via `the Adafruit Blinka library <https://github.com/adafruit/Adafruit_Blinka>`_. You will need to separately install the CircuitPython library for each sensor you want to scrape. For each sensor, you will then need to manually create and configure an instance of GenericSensor.

Example for `the BH1750 light sensor <https://github.com/adafruit/Adafruit_CircuitPython_BH1750>`_: ::

    # install the sensor library
    pip3 install adafruit-circuitpython-bh1750

    # create the sensor object
    import board
    i2c = board.I2C()
    bh1750 = adafruit_bh1750.BH1750(i2c)

    # prepare to poll / scrape
    GenericSensor(bh1750)

Although GenericSensor will only scrape numerical values, these aren't always wanted. For example, `the Adafruit SCD30 sensor class exposes an unwanted self_calibration_enabled property <https://github.com/adafruit/Adafruit_CircuitPython_SCD30/blob/b3d9bd141ae86ec4f871ae42a35d208003672c02/adafruit_scd30.py#L130>`_. You can use :mod:`Stream <snsary.streams.stream>` filters to clean up the output e.g.::

    GenericSensor(scd30).stream.filter_names('CO2', 'temperature', 'relative_humidity')

Warning: the Adafruit I2C library does have `a locking feature <https://github.com/adafruit/Adafruit_Blinka/blob/fa80f7d2ef51b0aea92196e96c3584512d30e64d/src/adafruit_blinka/__init__.py#L61-L66>`_, but it's not thread-safe. This means it's possible for multiple sensors to conflict if they poll for data around the exact same time.
"""

from snsary.models import Reading
from snsary.sources import PollingSensor
from snsary.utils import scraper


class GenericSensor(PollingSensor):
    def __init__(
        self,
        device,
        period_seconds=10,
    ):
        PollingSensor.__init__(self, period_seconds=period_seconds)
        self.__device = device
        self.__scraper = scraper.for_class(type(device))

    @property
    def device(self):
        return self.__device

    @property
    def name(self):
        return type(self.device).__name__

    def ready(self, **kwargs):
        """
        Returns ``True`` if the device is ready to be sampled, based on the kwargs to ``sample``. The decision to sample could equally be made based on the ``device`` itself. The default implementation returns ``True``.
        """
        return True

    def sample(self, **kwargs):
        if not self.ready(**kwargs):
            self.logger.info("Still warming up, no data yet.")
            return []

        return [
            Reading(
                sensor_name=self.name,
                name=name,
                value=value,
                timestamp=kwargs["timestamp"],
            )
            for (name, value) in self.__scraper(self.device)
        ]



"""
An Output has a ``publish`` method to receive :mod:`Readings <snsary.models.reading>`. Note that ``publish`` may be called concurrently if an Output is attached to multiple asynchronous :mod:`Sources <snsary.sources.source>`.
"""

from .batch_output import BatchOutput
from .mock_output import MockOutput
from .output import Output

__all__ = [
    "BatchOutput",
    "MockOutput",
    "Output",
]



from .tracker import Tracker


class MaxTracker(Tracker):
    def __init__(self, id, *, names, on_change):
        self.on_change = on_change
        Tracker.__init__(self, id, **{name: max for name in names})



"""
Computes the mean, max, min and median (p50) of :mod:`Readings <snsary.models.reading>` over consecutive windows. The name of each computation is appended to the name of the :mod:`Reading <snsary.models.reading>` e.g. ``myreading--mean``.
"""

from statistics import mean, median

from .window import Window


class WindowSummary(Window):
    def aggregate(self, readings):
        def __dup_reading(name, value):
            base_reading = readings[-1]

            return base_reading.dup(
                name=base_reading.name + f"--{name}",
                value=value,
            )

        values = [r.value for r in readings]

        return [
            __dup_reading("mean", mean(values)),
            __dup_reading("max", max(values)),
            __dup_reading("min", min(values)),
            __dup_reading("p50", median(values)),
        ]



"""
A function accepts a single :mod:`Reading <snsary.models.reading>` as an argument and can return zero or more :mod:`Readings <snsary.models.reading>`. This module contains classes whose instances are callable and act as functions.
"""

from .filter import Filter
from .function import Function
from .rename import Rename
from .window import Window
from .window_average import WindowAverage
from .window_summary import WindowSummary

__all__ = [
    "Filter",
    "Window",
    "WindowAverage",
    "Function",
    "WindowSummary",
    "Rename",
]



from datetime import datetime
from threading import Event, Thread

from snsary import system


class Poller(system.Service):
    def __init__(self, *, period_seconds):
        system.Service.__init__(self)
        self.__period = period_seconds
        self.__stop = Event()
        self.__thread = Thread(target=self.__loop, daemon=True)

    @property
    def period(self):
        return self.__period

    def start(self):
        self.__start_time = datetime.now().astimezone()
        self.__thread.start()

    def stop(self):
        self.__stop.set()
        self.__thread.join()

    def __loop(self):
        while not self.__stop.is_set():
            now = datetime.now().astimezone()

            try:
                self.tick(**self.__tick_kwargs(now))
            except Exception as e:
                self.logger.exception(e)

            now2 = datetime.now().astimezone()
            delay = int((now2 - now).total_seconds())

            if delay > self.period:
                self.logger.warning(f"Took too long to get sample: {delay}s.")
            else:
                self.__stop.wait(timeout=(self.period - delay))

    def tick(self, **kwargs):
        pass

    def __tick_kwargs(self, now):
        return {
            "now": now,
            "start_time": self.__start_time,
            "timestamp": int(now.timestamp()),
            "elapsed_seconds": int((now - self.__start_time).total_seconds()),
        }



from ..logging import get_logger
from .history import History
from .sample import Sample


class Monitor:
    """
    Analyses :mod:`Samples <snsary.utils.tracing.sample>` of the execution of some code.
    """

    def analyse(self, sample):
        raise NotImplementedError()

    @classmethod
    def factory(cls, **kwargs):
        return lambda: cls(**kwargs)


class TimeSeriesAlert(Monitor):
    """
    Provides a :mod:`History <snsary.utils.tracing.history>` of :mod:`Samples <snsary.utils.tracing.sample>` over time to support time series alerting. Each subclass is responsible for managing the history.
    """

    MAX_HISTORY = 300

    def __init__(self, *, history=None):
        self._history = (
            history
            if history is not None
            else History(
                max_length=self.MAX_HISTORY,
            )
        )


class LivenessAlert(TimeSeriesAlert):
    """
    Logs an error when a complete history of samples is nothing but failures.

    The history is reset every time the monitor logs an error.
    """

    def analyse(self, sample):
        self._history.add(sample)

        if len(self._history) < self._history.max_length:
            return

        for sample in self._history:
            if sample == Sample.SUCCESS:
                return

        get_logger().error(
            f"Alert: {self._history.max_length} failures in sample window"
        )
        self._history.reset()


class GapAlert(TimeSeriesAlert):
    """
    Logs an error when there are too many gaps in ``SUCCESS`` samples. A gap
    is one or more ``FAILURE`` samples followed by a ``SUCCESS`` sample.

    The history is reset every time the monitor logs an error.
    """

    MAX_GAPS = 30

    def __init__(self, *, max_gaps=MAX_GAPS, **kwargs):
        TimeSeriesAlert.__init__(self, **kwargs)
        self.__max_gaps = max_gaps

    def analyse(self, sample):
        self._history.add(sample)

        if self.__count_gaps() <= self.__max_gaps:
            return

        get_logger().error(f"Alert: {self.__max_gaps + 1} gaps in last sample window")
        self._history.reset()

    def __count_gaps(self):
        gaps = 0
        previous = None

        for sample in self._history:
            if previous == Sample.FAILURE and sample == Sample.SUCCESS:
                gaps += 1

            previous = sample

        return gaps



"""
Path-based key / value store with optional defaults.

Basic key / value access looks like this: ::

    config = Config()
    config.set("some.unique.path.attribute", 1)
    config.get("some.unique.path.attribute")  # => 1

Defaults can be set for each key access: ::

    config.get("some.other.path.attribute", default=2)  # => 2

Attribute defaults can also be set globally: ::

    config.set("attribute", 3)
    config.get("some.other.path.attribute", default=2)  # => 3
"""


class Config:
    def __init__(self):
        self.reset()

    def get(self, path, default=None):
        if path in self.__backend:
            return self.__backend[path]

        prefix, _, attribute = path.rpartition(".")

        if attribute in self.__backend:
            return self.__backend[attribute]

        if default is not None:
            return default

        raise KeyError(f"No config defined for '{path}'")

    def set(self, path, value):
        self.__backend[path] = value

    def reset(self):
        self.__backend = {}



"""
Outputs KWh (electricity) or cubic meter (gas) consumption from a specified Smart Meter in `half hour intervals <https://developer.octopus.energy/docs/api/#consumption>`_. At the time of writing, data is only available for the previous day sometime on the following day (not sure exactly when). Create instances with ``.<fuel_type>_from_env()``.

- ``.electricity_from_env()`` expects:
    - OCTOPUS_ELECTRICITY_MPAN
    - OCTOPUS_ELECTRICITY_SERIAL
    - OCTOPUS_TOKEN

- ``.gas_from_env()`` expects:
    - OCTOPUS_GAS_MPRN
    - OCTOPUS_GAS_SERIAL
    - OCTOPUS_TOKEN

"""

import os
from datetime import timedelta

import pyrfc3339

from snsary.models import Reading
from snsary.sources import PollingSensor
from snsary.utils import request


class OctopusSensor(PollingSensor):
    CONSUMPTION_URL = "https://api.octopus.energy/v1/{fuel_type}-meter-points/{mpxn}/meters/{serial_number}/consumption/?period_from={period_from}&order_by=period"

    @classmethod
    def electricity_from_env(cls):
        return cls(
            mpxn=os.environ["OCTOPUS_ELECTRICITY_MPAN"],
            serial_number=os.environ["OCTOPUS_ELECTRICITY_SERIAL"],
            token=os.environ["OCTOPUS_TOKEN"],
            fuel_type="electricity",
        )

    @classmethod
    def gas_from_env(cls):
        return cls(
            mpxn=os.environ["OCTOPUS_GAS_MPRN"],
            serial_number=os.environ["OCTOPUS_GAS_SERIAL"],
            token=os.environ["OCTOPUS_TOKEN"],
            fuel_type="gas",
        )

    def __init__(self, *, mpxn, serial_number, token, fuel_type):
        PollingSensor.__init__(
            self,
            period_seconds=30 * 60,  # 30 mins
        )

        self.__token = token
        self.__mpxn = mpxn
        self.__serial_number = serial_number
        self.__fuel_type = fuel_type

    @property
    def name(self):
        return f"octopus.{self.__fuel_type}"

    def sample(self, now, **kwargs):
        start = now - timedelta(days=1)
        start = start.replace(hour=0, minute=0, second=0, microsecond=0)

        url = self.CONSUMPTION_URL.format(
            mpxn=self.__mpxn,
            serial_number=self.__serial_number,
            period_from=pyrfc3339.generate(start),
            fuel_type=self.__fuel_type,
        )

        self.logger.debug(f"Request {url}")
        response = request.retrying_session().get(
            url,
            auth=(self.__token, ""),
        )
        response.raise_for_status()

        samples = response.json()["results"]
        self.logger.debug(f"Response {samples}")

        for sample in samples:
            yield self.__reading_from_sample(sample)

    def __reading_from_sample(self, sample):
        sample_datetime = pyrfc3339.parse(sample["interval_end"])
        sample_timestamp = int(sample_datetime.timestamp())

        return Reading(
            sensor_name=self.name,
            name="consumption",
            timestamp=sample_timestamp,
            value=sample["consumption"],
        )



"""
An adapter for `the PyPMS library <https://github.com/avaldebe/PyPMS>`_.

Example of creating an instance::

    PyPMSSensor(sensor_name='PMSx003')
"""
import dataclasses

from pms.core.reader import SensorReader

from snsary.models import Reading
from snsary.sources import PollingSensor


class PyPMSSensor(PollingSensor):
    """
    ``warm_up_seconds`` is necessary for some sensors e.g. for the PMSA003 the first two samples always `raise an InconsistentObservation exception <https://github.com/avaldebe/PyPMS/blob/04ff8edede7d780018cd00a7fcf78ffed43c0de4/src/pms/sensors/plantower/pmsx003.py#L63>`_.
    """

    class SnsaryReader(SensorReader):
        """
        PyPMS SensorReader customisations.
        """

        def _pre_heat(self):
            """
            Disable pre-heat to avoid blocking samples.

            Sensors will raise an exception if they are not ready.
            """
            pass

    def __init__(
        self,
        *,
        sensor_name,
        port="/dev/ttyS0",
        timeout=5,
    ):
        self.__reader = self.SnsaryReader(
            sensor=sensor_name,
            port=port,
            samples=1,
            max_retries=0,
            timeout=timeout,
        )

        PollingSensor.__init__(
            self,
            period_seconds=10,
        )

    @property
    def name(self):
        return self.__reader.sensor.name

    def start(self):
        self.__reader.open()
        PollingSensor.start(self)

    def stop(self):
        PollingSensor.stop(self)

        try:
            self.__reader.close()
        except Exception as e:
            self.logger.exception(e)

    def sample(self, timestamp, elapsed_seconds, **kwargs):
        if elapsed_seconds < self.__reader.pre_heat:
            self.logger.info("Still warming up, no data yet.")
            return []

        obs = next(self.__reader())

        return [
            Reading(
                sensor_name=self.name,
                name=key,
                value=value,
                timestamp=timestamp,
            )
            for key, value in dataclasses.asdict(obs).items()
            if key not in ("time")
        ]



from .stream import Stream


class SimpleStream(Stream):
    def __init__(self):
        self.__outputs = []

    def publish(self, reading):
        for output in self.__outputs:
            try:
                output.publish(reading)
            except Exception as e:
                output.logger.exception(e)

    def subscribe(self, output):
        self.__outputs += [output]



"""
A store has a dict-like interface to get and set data that a consumer wants to persist between program restarts e.g. the current :mod:`Readings <snsary.models.reading>` in a :mod:`Window <snsary.functions.window>`. This module provides a ``get_storage`` function and a ``HasStore`` trait to make it easy access to persistent storage.

By default persistent storage is disabled to avoid creating files without warning - the data stored is only held in memory and not written to disk. Set the ``STORAGE_PATH`` environment variable to enable persistent storage to a file e.g. ``export STORAGE_PATH=snsary.db``.

Storage trackers
================

Trackers can be used to keep track of values of series of :mod:`Readings <snsary.models.reading>` with specified names using persistent storage. New values for each series are individually compared with old values using the specified function for the series e.g. ::

    Tracker('id-for-persistent-storage', myreading=max, othername=min)

Call ``update`` when new readings are available for any of the series. ``on_change`` is then called with :mod:`Reading <snsary.models.reading>` kwargs ``old`` and ``new`` when each of the tracked series changes - set to a different function to subscribe to changes.

Storage internals
=================

Persistent storage uses the `pickle <https://docs.python.org/3/library/pickle.html>`_ module internally. Data to be stored is only written to disk when a program exits, to prevent any issues from concurrent writes. If the specified file doesn't exist when a program starts, it will be created when data is first written.

Even persistent storage will expire, to prevent a build up of dead data over time as a program evolves with different consumers of storage and different data to store. Set the ``STORAGE_TTL`` environment variable to override the default TTL of 1 day.

The storage format is not versioned and may break without warning after an upgrade. You will need to manually remove the file at ``STORAGE_PATH`` if this happens.
"""

from .backend import HasStore, get_storage
from .max_tracker import MaxTracker
from .null_tracker import NullTracker
from .tracker import Tracker

__all__ = [
    "HasStore",
    "get_storage",
    "MaxTracker",
    "NullTracker",
    "Tracker",
]



"""
Sends batches of :mod:`Readings <snsary.models.reading>` as "points" to a specified InfluxDB bucket at a specified endpoint e.g. `InfluxDB Cloud <https://www.influxdata.com/products/influxdb-cloud>`_. Each point is named after the Reading and tagged by **sensor** and **host**.

Create an instance with ``.from_env()``, which expects:

- INFLUXDB_URL
- INFLUXDB_TOKEN
- INFLUXDB_ORG
- INFLUXDB_BUCKET
"""

import os
import platform

from influxdb_client import InfluxDBClient, Point
from influxdb_client.client.write_api import SYNCHRONOUS

from snsary.outputs import BatchOutput


class InfluxDBOutput(BatchOutput):
    @classmethod
    def from_env(cls):
        return cls(
            url=os.environ["INFLUXDB_URL"],
            token=os.environ["INFLUXDB_TOKEN"],
            org=os.environ["INFLUXDB_ORG"],
            bucket=os.environ["INFLUXDB_BUCKET"],
        )

    def __init__(self, *, url, token, org, bucket):
        BatchOutput.__init__(self)
        client = InfluxDBClient(url=url, token=token, org=org)
        self.__bucket = bucket
        self.__write_api = client.write_api(write_options=SYNCHRONOUS)

    def publish_batch(self, readings):
        points = [
            Point(reading.name)
            .tag("sensor", reading.sensor_name)
            .tag("host", platform.node())
            .field("value", reading.value)
            .time(reading.timestamp, write_precision="s")
            for reading in readings
        ]

        lines = [point.to_line_protocol() for point in points]
        self.logger.debug("Sending " + str(lines))

        self.__write_api.write(bucket=self.__bucket, record=points)



"""
Computes a mean average of :mod:`Readings <snsary.models.reading>` over fixed, consecutive windows.
"""

from statistics import mean

from .window import Window


class WindowAverage(Window):
    def aggregate(self, readings):
        aggregate_value = mean(r.value for r in readings)
        return [readings[-1].dup(value=aggregate_value)]



import signal
from threading import Event, Thread

from snsary.utils import logging

from .service import Service, get_services

_logger = logging.get_logger()


def start():
    for service in get_services():
        service.logger.debug("Starting.")
        service.start()

    _logger.info("Started.")


def start_and_wait():
    start()
    wait()


def stop(*_):
    _logger.info("Stopping.")

    for service in get_services():
        __stop_service(service)

    _logger.info("Bye.")


def wait(*, handle_signals=True):
    if handle_signals:
        signal.signal(signal.SIGINT, stop)
        signal.signal(signal.SIGTERM, stop)

    class Waiter(Service):
        def __init__(self):
            Service.__init__(self)
            self.__event = Event()

        def wait(self):
            self.__event.wait()

        def stop(self):
            self.__event.set()

    Waiter().wait()


def __stop_service(service):
    service.logger.debug("Stopping.")
    thread = Thread(target=service.stop, daemon=True)
    thread.start()

    thread.join(timeout=1)
    service.logger.debug("Stopped.")

    if thread.is_alive():
        service.logger.error("Failed to stop.")



import requests
from urllib3 import util


def retrying_session():
    adapter = requests.adapters.HTTPAdapter(
        max_retries=util.Retry(
            total=3,
            allowed_methods=False,
        )
    )

    session = requests.Session()
    session.mount("http://", adapter)
    session.mount("https://", adapter)
    return session



"""
Manually subscribing each :mod:`Source <snsary.sources.source>` to each :mod:`Output <snsary.outputs.output>` is repetitive, especially when there are multiple Outputs. MultiSource combines multiple Sources as one. Just like a :mod:`Sensor <snsary.sources.sensor>`, a MultiSource also exposes a stream to make it easier to work with: ::

    MultiSource(MockSensor(), MockSensor()).stream.into(MockOutput())
"""

from .source import Source


class MultiSource(Source):
    def __init__(self, *sources):
        from snsary.streams import AsyncStream

        self.__stream = AsyncStream()

        for source in sources:
            source.subscribe(self.__stream)

    def subscribe(self, output):
        self.stream.subscribe(output)

    @property
    def stream(self):
        return self.__stream



"""
Sends batches of :mod:`Readings <snsary.models.reading>` as "mutations" to a specified `DataStax Astra DB <https://docs.datastax.com/en/astra/docs/index.html>`_ GraphQL endpoint using the `Python GQL client <https://github.com/graphql-python/gql>`_, as a means of inserting into a Cassandra table.

GraphQL replaces normal REST with a server-defined language / API in nested {} format. `DataStax generates the API automatically based on the keyspaces and tables that exist <https://docs.datastax.com/en/astra/docs/using-the-astra-graphql-api.html>`_.

GraphQL isn't well suited to timeseries data: every insertion ("mutation") must have a unique alias. The output compensates for this by using throwaway "r0", "r1" aliases for each reading mutation. See the tests for an example. In order to simplify building each request, the output makes an initial request to get the schema for the keyspace in order to utilise the DSL feature.

Create an instance with ``.from_env()``, which expects:

- DATASTAX_URL
- DATASTAX_TOKEN (needs API write permission)

Setting up Astra DB
===================

Create a DB table in Astra as follows: ::

    CREATE TABLE reading (
        host text,
        sensor text,
        metric text,
        timestamp timestamp,
        value double,
        PRIMARY KEY ((host,sensor,metric), timestamp)
    )
    WITH CLUSTERING ORDER BY (timestamp DESC)
    AND default_time_to_live = 33696000;

The output specifies a TTL for each insertion, which may be configurable in future. Having a default TTL for the table is optional but reduces the risk of data remaining indefinitely. Note that it's current not possible to specify other options like the compaction strategy. However, `the DataStax docs say the default strategy is suitable for time series data <https://docs.datastax.com/en/astra/docs/datastax-astra-database-limits.html>`_.

Querying the data
=================

Example GraphQL query for data in the table: ::

    query readings {
      data:reading (
        options: {
          pageSize: 10000, limit: 10000
        },
        filter: {
          host: { eq: "raspberrypi" },
          sensor: { eq: "PMSx003" },
          metric: { eq: "pm25--max/minute" },
          timestamp: { gt: "${__from:date:iso}", lt: "${__to:date:iso}" }
        }
      ) {
        values {
          host, sensor, metric, timestamp, value
        }
      }
    }

Note that the "$" variables are `Grafana global variables <https://grafana.com/docs/grafana/latest/variables/variable-types/global-variables/>`_.
"""

import logging
import os
import platform

import pytz
from gql import Client
from gql.dsl import DSLMutation, DSLSchema, dsl_gql
from gql.transport.requests import RequestsHTTPTransport
from graphql import print_ast

from snsary.outputs import BatchOutput


class GraphQLOutput(BatchOutput):
    TTL = 33696000  # 13 months

    def __init__(self, url, token):
        self.__client = Client(
            transport=RequestsHTTPTransport(
                url=url,
                headers={"X-Cassandra-Token": token},
            ),
        )

        # logs per request otherwise
        logging.getLogger("gql").setLevel(logging.WARNING)

        BatchOutput.__init__(self)
        self.__init_schema()

    @classmethod
    def from_env(cls):
        return cls(
            url=os.environ["DATASTAX_URL"],
            token=os.environ["DATASTAX_TOKEN"],
        )

    def publish_batch(self, readings):
        mutations = list(
            self.__mutation(reading).alias(f"r{i}")
            for i, reading in enumerate(readings)
        )

        query = dsl_gql(DSLMutation(*mutations))
        self.logger.debug(f"Sending {print_ast(query)}")
        self.__client.execute(query)

    def __mutation(self, reading):
        value = {
            # using UTC ensures test stability when run in different zones
            "timestamp": reading.datetime.astimezone(pytz.utc).isoformat(),
            "sensor": reading.sensor_name,
            "host": platform.node(),
            "metric": reading.name,
            "value": float(reading.value),
        }

        throwaway_result = self.__dsl.readingMutationResult.value.select(
            self.__dsl.reading.metric
        )

        return self.__dsl.Mutation.insertreading.args(
            options={"ttl": self.TTL}, value=value
        ).select(throwaway_result)

    def __init_schema(self):
        with self.__client as session:
            session.fetch_schema()

        self.logger.debug("Initialised schema.")
        self.__dsl = DSLSchema(self.__client.schema)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: test.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\ntest.proto\x12\x06snsary"Y\n\x07Reading\x12\x0c\n\x04host\x18\x01 \x02(\t\x12\x0e\n\x06sensor\x18\x02 \x02(\t\x12\x0e\n\x06metric\x18\x03 \x02(\t\x12\x11\n\ttimestamp\x18\x04 \x02(\t\x12\r\n\x05value\x18\x05 \x02(\x01'
)


_READING = DESCRIPTOR.message_types_by_name["Reading"]
Reading = _reflection.GeneratedProtocolMessageType(
    "Reading",
    (_message.Message,),
    {
        "DESCRIPTOR": _READING,
        "__module__": "test_pb2"
        # @@protoc_insertion_point(class_scope:snsary.Reading)
    },
)
_sym_db.RegisterMessage(Reading)

if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _READING._serialized_start = 22
    _READING._serialized_end = 111
# @@protoc_insertion_point(module_scope)
